// $Id: ParameterMap.h 1147 2020-10-12 12:53:31Z ge $
/// \file ParameterMap.h
/// \brief This is a generic class that handles parameters in form XX:YY
///
/// We need to handle parameters frequently. Here we are able to store and
/// retrieve parameters of type XX:YY ZZ, XX is a general designation such as AT,
/// YY is the actual parameter name and Z its value. We can have more than one
/// general designation XX1:XX2:XX3:YY which expands as XX1:YY XX2:YY XX3:YY.
/// Also we can use * in place of XX meaning that any parameter request without
/// a general_class XX will be assigned this value.
/// $Revision: 1147 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_PARAMETERMAP_H
#define GBC_EXP_PARAMETERMAP_H "$Id: ParameterMap.h 1147 2020-10-12 12:53:31Z ge $"
#include <map>
#include <deque>
#include <valarray>
#include <iostream>
#include <fstream>
#include <sstream>
#include "MathAux.h"
#include<boost/algorithm/string/split.hpp>                                      
#include<boost/algorithm/string.hpp>                                            
#include <boost/regex.hpp>
 
namespace gbc
  {
  template<class _Tp=double>
  /// \brief This class handles parameters used by Potentials and retrieves them from files.
  ///
  /// The parameters are identified by an ID usually constructed from a parameter name
  /// and an object ID. For example the parameter D of the morse potential is named
  /// morse.D and if used for an AT base pair it appears as AT:morse.D
  /// The parameter file may define a default ID "*", which matches any parameter
  /// which has not been identified previsouly. 
  class ParameterMap: public std::map<std::string,_Tp>
    {
    public:
      typedef _Tp                                        parameter_type;
      typedef std::string                                identifier_type;
      typedef std::map<identifier_type,parameter_type>   map_type;
      typedef std::deque<identifier_type>                field_deque_type;  ///< List of identifications
      typedef std::deque<parameter_type>                 parameter_deque_type;
      typedef std::map<std::string,std::string>          equivalence_map_type;
      typedef std::deque<std::string>                    string_deque_type;

    static const char Field_separator=':'; ///< This is how we separate each field.
    static const char Default_sub_id='*';  ///< The default sub-id.

    bool Record_variation_differences;
    identifier_type Identify;
    parameter_deque_type Randomization_differences;
    map_type Temporary_storage;
    
    ParameterMap(void): Record_variation_differences(false) {}

    /// \brief Sets the parameter identified by its ID.
    ///
    /// If the parameter cannot be found by its ID we search for a default
    /// which starts by "*:"
    /// ATTENTION: true if parameter was **modified**
    inline bool get_parameter(const identifier_type id,  ///< ID or list of IDs
                                parameter_type& par,     ///< The parameter that is modified
                                bool complain=true       ///< Complains if not found
                             ) const 
      {
      bool modified=false;
      find_parameter(id,par,modified,complain);
      return modified;
      }
     
    /// \brief Sets the parameter identified by its ID.
    ///
    /// If the parameter cannot be found by its ID we search for a default
    /// which starts by "*:"
    /// ATTENTION: true if parameter was found
    inline bool find_parameter(const identifier_type id,  ///< ID or list of IDs
                                parameter_type& par,     ///< The parameter that is modified
                                bool &modified,          ///< This is what differentiates this function from get_parameter 
                                bool complain=true       ///< Complains if not found
                             ) const 
      {
      typename map_type::const_iterator it=this->find(id);
      if (it != this->end()) 
        {
        modified = par != it->second;
        par=it->second;
        return true;
        }
      else 
        {
        // For the case where several base pairs, or nearest neighbour pairs are
        // assigned the several parameter value in the form AT:TA:par we first separate
        // the fields AT, TA, par with the function separate.
        // sf[0] is the name of the parameter (par in our example).
        field_deque_type sf=separate(id);
        if (sf.size() > (size_t)0)
          {
          if (sf.size() >= 2) it=this->find(sf[1]+Field_separator+sf[0]); // we try the first in the list
          else it=this->end(); //if only one element exist in sf then we should use a generic parameter
          if (it == this->end()) it=this->find("*:"+sf[0]); //we try a generic parameter

          if (it == this->end()) // ATTENTION: experimental and currently works only for NN
            {
            // perl -Mre=debug -e '"TT^j_TT^j" =~ /^(.+?)(|\^.+)\_(.+?)(|\^.+)$/; print "res=[$1] [$3]";'
            boost::regex match("^(.+?)(|\\^.+)\\_(.+?)(|\\^.+)$");
            std::string rep=boost::regex_replace(sf[1],match,"\\1_\\3");
            if (rep != sf[1]) it=this->find(rep+Field_separator+sf[0]);
            if (it != this->end())
              {
              std::cout << "Warning: using " << rep+Field_separator+sf[0] << " for " << sf[1]+Field_separator+sf[0] << std::endl;              
              }
            }
            
          if (it != this->end()) 
            {
            modified = par != it->second;
            par=it->second;
            return true;
            }
          else 
            {
            if (complain)
              {
              std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Error: generic parameter()" << "*:"+sf[0] << " for " << id << " not found." << std::endl;
              std::cerr << "\nWe currently have these parameters in our map:\n" << *this << "\nProgram terminating\n";
              exit(1);
              }
            return false;
            }
          }
        }
     if (complain)
       {
       std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Warning: parameter()" << id << " not found." << std::endl;
       std::cerr << "\nWe currently have these parameters in our map:\n" << *this << "\nProgram terminating\n";
       exit(1);
       }
     return false;
     }



    /// \brief Gets the parameter identified by its ID.
    ///
    /// If the parameter cannot be found by its ID we search for a default
    /// which starts by "*:"
    /// \Returns: parameter 
    inline parameter_type  get_parameter_or_generic(const identifier_type id,  ///< ID or list of IDs
                                bool complain=true       ///< Complains if not found
                             ) const 
      {
      typename map_type::const_iterator it=this->find(id);
      if (it != this->end()) 
        {
        return it->second;
        }
      else 
        {
        // For the case where several base pairs, or nearest neighbour pairs are
        // assigned the several parameter value in the form AT:TA:par we first separate
        // the fields AT, TA, par with the function separate.
        // sf[0] is the name of the parameter (par in our example).
        if (complain)
          {
          std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Warning: parameter() for " << id << " not found. Trying generic parameters instead." << std::endl;
          std::cerr << *this;
          }
        field_deque_type sf=separate(id);
        if (sf.size() > (size_t)0)
          {
          if (sf.size() >= 2) it=this->find(sf[1]+Field_separator+sf[0]); // we try the first in the list
          else it=this->end(); //if only one element exist in sf then we should use a generic parameter
          if (it == this->end()) it=this->find("*:"+sf[0]); //we try a generic parameter
          if (it != this->end()) 
            {
            return it->second;
            }
          else 
            {
            if (complain)
              {
              std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Warning: generic parameter()" << "*:"+sf[0] << " for " << id << " not found." << std::endl;
              std::cerr << *this;
              }
            return parameter_type();
            }
          }
        }
     return parameter_type();
     }     
     
    /// Sets a new value par in the existing parameter in map
    inline bool set_parameters(const identifier_type id,  ///< ID or list of IDs
                               const parameter_type par) ///< The value of the parameter that we should set
      {
      typename map_type::iterator it=this->find(id);
      if (it != this->end()) 
        {
        it->second=par;
        return true;
        }
      else 
        {
        // For the case where several base pairs, or nearest neighbour pairs are
        // assigned the several parameter value in the form AT:TA:par we first separate
        // the fields AT, TA, par with the function separate.
        // sf[0] is the name of the parameter (par in our example).
        field_deque_type sf=separate(id);
        if (sf.size() > (size_t)0)
          {
          for (size_t n=1; n < sf.size(); ++n)
            {
            it=this->find(sf[n]+Field_separator+sf[0]);
            if (it != this->end()) it->second=par;
            else insert_parameter(sf[n]+Field_separator+sf[0],par);
            }
          return true;
          }
        else return false;
        }
     std::cerr << "Warning set_parameters(): parameter " << id << " not found." << std::endl;
     return false;
     }

    /// \brief Creates ID with empty parameter
    inline void insert_id(identifier_type id)
      {
      this->insert(typename map_type::value_type(id,parameter_type()));
      }

    /// \brief Inserts ID and its parameter
    inline void insert_parameter(identifier_type id, const parameter_type& par, std::string wmsg=std::string())
      {
      if (this->find(id) != this->end())
        {
        if (this->find(id)->second != par)
          {
          if (wmsg.size()) std::cout << "Warning:replacing " << id 
            << " old value=" << this->find(id)->second
            << " new value=" << par << " (" << wmsg << ")" << std::endl;
          this->find(id)->second=par;
          }
        }
      else this->insert(typename map_type::value_type(id,par));
      }

    /// \brief Randomises the parameters by a certain amount
    inline void randomise(parameter_type rnd) ///< Average deviation
      {
      typename map_type::iterator it;
      if (Temporary_storage.size() > 0) restore_original();
      else Temporary_storage=static_cast<map_type>(*this);
      for(it=this->begin(); it != this->end(); ++it)
        {
        parameter_type temp=it->second;
        double dice=gauss_ran(rnd);
        it->second *= (1.0+dice);
        if (temp != parameter_type() && Record_variation_differences)
          Randomization_differences.push_back((temp-it->second)/temp);
        }
      }

    inline void restore_original(void)
      {
      static_cast<map_type>(*this)=Temporary_storage;
      }

    /// \brief Prints a list of IDs and parameters
    friend std::ostream& operator<<(std::ostream& out,const ParameterMap& pm)
      {
      typename map_type::const_iterator it;
      for(it=pm.begin(); it != pm.end(); ++it)
        {
        out << it->first << " " << it->second << std::endl;
        }
      return out;
      }

    /// \brief Prints a list of IDs and parameters
    friend bool operator==(const ParameterMap& pm1, const ParameterMap& pm2)
      {
      if (pm1.size() != pm2.size()) return false;
        else                        return std::equal(pm1.begin(), pm1.end(), pm2.begin());
      }

    /// \brief Splits the field into several components separated by colon.
    ///
    /// This method first gets the parameter name, and afterwards its sub-id.
    inline field_deque_type separate(std::string field) const
      {
      field_deque_type extracted;

      std::string::size_type pos=field.find(Field_separator);
      if (pos != std::string::npos)
        {
        boost::algorithm::split(extracted, field, boost::is_any_of(":"),boost::token_compress_on);
        //here we put the last item to front
        extracted.push_front(extracted.back());extracted.pop_back();
        }
      else 
        {
        boost::algorithm::split(extracted, field, boost::is_any_of("+=, "),boost::token_compress_on);
        }
      //std::copy(extracted.begin(),extracted.end(),std::ostream_iterator<std::string>(std::cout," "));
      return extracted; 
      };
      
    inline void read_from_list_of_files(string_deque_type &par_file_deque)
      {
      typename string_deque_type::iterator pit;
      for(pit=par_file_deque.begin(); pit != par_file_deque.end(); pit++)
        read_from(*pit);
      }

    inline void read_from(std::string fname)
      {
      std::ifstream parfile; parfile.open(fname.c_str());
      if (!parfile.good()) 
        {
        std::cerr << __FILE__ << ":" << __LINE__ << " Could not read " << fname << std::endl;
        exit(1);
        }
      parfile >> Identify; //Reads the first line of parfile which is ignored subsequently
      while (!parfile.eof())
        {
        char line[1024];
        parfile.getline(line,1024);
        if (parfile.gcount() > 1 && (line[0] != '+')) 
          {
          std::string field;
          std::stringstream sline;
          sline.str(line);
          sline >> field;
          double par;
          sline >> par;
          field_deque_type sf=separate(field);
          if (sf.size() == (size_t)1) 
            {
            insert_parameter(sf[0],par,fname);
            }
          else
            {
            for (size_t n=1; n < sf.size(); ++n)
            insert_parameter(sf[n]+Field_separator+sf[0],par,fname);
            }
          }
        } 
      parfile.close();
      }
      
    inline parameter_type get_if_exists(std::string id)
      {
      typename map_type::const_iterator it=this->find(id);
      if (it != this->end()) 
        {
        return it->second;
        }
      return parameter_type();
      }
 
    /// \brief Generates a set or multiset of all id in use
    ///
    /// Specify the _Container either as set or multiset. A set
    /// will give you unique ids, multiset will give you repeated ids.
    template<class _Container>
    inline void get_all_id(_Container& collection) const
      {
      field_deque_type first_field;
      typename map_type::const_iterator it;
      for(it=this->begin(); it != this->end(); ++it)
        {
        first_field=separate(it->first);
        for(size_t n=1; n < first_field.size(); ++n)
          collection.insert(first_field[n]);
        }
      }

    };
};
#endif
